part of cryptography;

/// 循环冗余校验（Cyclic Redundancy Check，CRC）
abstract class CRC {
  /// 名称
  String get name;

  /// 宽度，最大支持 64 位
  int get width;

  /// 多项式
  int get poly;

  /// 初始值
  int get init;

  /// 输入反转
  bool get refIn;

  /// 输出反转
  bool get refOut;

  /// 输出异或值
  int get xorOut;

  /// 使用指定参数创建 CRC 实例
  ///
  /// [mapping] - 指定是否使用查表法进行计算
  factory CRC(String name, int width, int poly, int init, bool refIn,
      bool refOut, int xorOut,
      [bool mapping = false]) {
    final crc = mapping
        ? _MappingCRC(name, width, poly, init, refIn, refOut, xorOut)
        : _CRC(name, width, poly, init, refIn, refOut, xorOut);
    return crc;
  }

  /// CRC-4/ITU
  ///
  /// [mapping] - 指定是否使用查表法进行计算
  factory CRC.crc4ITU([bool mapping = false]) =>
      CRC('CRC-4/ITU', 4, 0x03, 0x00, true, true, 0x00, mapping);

  /// CRC-5/EPC
  ///
  /// [mapping] - 指定是否使用查表法进行计算
  factory CRC.crc5EPC([bool mapping = false]) =>
      CRC('CRC-5/EPC', 5, 0x09, 0x09, false, false, 0x00, mapping);

  /// CRC-5/ITU
  ///
  /// [mapping] - 指定是否使用查表法进行计算
  factory CRC.crc5ITU([bool mapping = false]) =>
      CRC('CRC-5/ITU', 5, 0X15, 0x00, true, true, 0x00, mapping);

  /// CRC-5/USB
  ///
  /// [mapping] - 指定是否使用查表法进行计算
  factory CRC.crc5USB([bool mapping = false]) =>
      CRC('CRC-5/USB', 5, 0x05, 0x1F, true, true, 0x1F, mapping);

  /// CRC-6/ITU
  ///
  /// [mapping] - 指定是否使用查表法进行计算
  factory CRC.crc6ITU([bool mapping = false]) =>
      CRC('CRC-6/ITU', 6, 0x03, 0x00, true, true, 0x00, mapping);

  /// CRC-7/MMC
  ///
  /// [mapping] - 指定是否使用查表法进行计算
  factory CRC.crc7MMC([bool mapping = false]) =>
      CRC('CRC-7/MMC', 7, 0x09, 0x00, false, false, 0x00, mapping);

  /// CRC-8
  ///
  /// [mapping] - 指定是否使用查表法进行计算
  factory CRC.crc8([bool mapping = false]) =>
      CRC('CRC-8', 8, 0x07, 0x00, false, false, 0x00, mapping);

  /// CRC-8/ITU
  ///
  /// [mapping] - 指定是否使用查表法进行计算
  factory CRC.crc8ITU([bool mapping = false]) =>
      CRC('CRC-8/ITU', 8, 0x07, 0x00, false, false, 0x55, mapping);

  /// CRC-8/MAXIM
  ///
  /// [mapping] - 指定是否使用查表法进行计算
  factory CRC.crc8MAXIM([bool mapping = false]) =>
      CRC('CRC-8/MAXIM', 8, 0x31, 0x00, true, true, 0x00, mapping);

  /// CRC-8/ROHC
  ///
  /// [mapping] - 指定是否使用查表法进行计算
  factory CRC.crc8ROHC([bool mapping = false]) =>
      CRC('CRC-8/ROHC', 8, 0x07, 0xFF, true, true, 0x00, mapping);

  /// CRC-16
  ///
  /// [mapping] - 指定是否使用查表法进行计算
  factory CRC.crc16([bool mapping = false]) =>
      CRC('CRC-16', 16, 0x8005, 0x0000, true, true, 0x0000, mapping);

  /// CRC-16/CCITT
  ///
  /// [mapping] - 指定是否使用查表法进行计算
  factory CRC.crc16CCITT([bool mapping = false]) =>
      CRC('CRC-16/CCITT', 16, 0x1021, 0x0000, true, true, 0x0000, mapping);

  /// CRC-16/CCITT-FALSE
  ///
  /// [mapping] - 指定是否使用查表法进行计算
  factory CRC.crc16CCITTFALSE([bool mapping = false]) => CRC(
      'CRC-16/CCITT-FALSE', 16, 0x1021, 0xFFFF, false, false, 0x0000, mapping);

  /// CRC-16/DNP
  ///
  /// [mapping] - 指定是否使用查表法进行计算
  factory CRC.crc16DNP([bool mapping = false]) =>
      CRC('CRC-16/DNP', 16, 0x3D65, 0x0000, true, true, 0xFFFF, mapping);

  /// CRC-16/MAXIM
  ///
  /// [mapping] - 指定是否使用查表法进行计算
  factory CRC.crc16MAXIM([bool mapping = false]) =>
      CRC('CRC-16/MAXIM', 16, 0x8005, 0x0000, true, true, 0xFFFF, mapping);

  /// CRC-16/MODBUS
  ///
  /// [mapping] - 指定是否使用查表法进行计算
  factory CRC.crc16MODBUS([bool mapping = false]) =>
      CRC('CRC-16/MODBUS', 16, 0x8005, 0xFFFF, true, true, 0x0000, mapping);

  /// CRC-16/USB
  ///
  /// [mapping] - 指定是否使用查表法进行计算
  factory CRC.crc16USB([bool mapping = false]) =>
      CRC('CRC-16/USB', 16, 0x8005, 0xFFFF, true, true, 0xFFFF, mapping);

  /// CRC-16/X25
  ///
  /// [mapping] - 指定是否使用查表法进行计算
  factory CRC.crc16X25([bool mapping = false]) =>
      CRC('CRC-16/X25', 16, 0x1021, 0xFFFF, true, true, 0xFFFF, mapping);

  /// CRC-16/XMODEM
  ///
  /// [mapping] - 指定是否使用查表法进行计算
  factory CRC.crc16XMODEM([bool mapping = false]) =>
      CRC('CRC-16/XMODEM', 16, 0x1021, 0x0000, false, false, 0x0000, mapping);

  /// CRC-32
  ///
  /// [mapping] - 指定是否使用查表法进行计算
  factory CRC.crc32([bool mapping = false]) => CRC(
      'CRC-32', 32, 0x04C11DB7, 0xFFFFFFFF, true, true, 0xFFFFFFFF, mapping);

  /// CRC-32/MPEG-2
  ///
  /// [mapping] - 指定是否使用查表法进行计算
  factory CRC.crc32MPEG2([bool mapping = false]) => CRC('CRC-32/MPEG-2', 32,
      0x04C11DB7, 0xFFFFFFFF, false, false, 0x00000000, mapping);

  /// 计算
  int calculate(List<int> data);

  /// 校验
  bool verify(List<int> data, int crc);
}

class _CRC implements CRC {
  final String _name;
  final int _width;
  final int _poly;
  final int _init;
  final bool _refIn;
  final bool _refOut;
  final int _xorOut;

  _CRC(this._name, this._width, this._poly, this._init, this._refIn,
      this._refOut, this._xorOut);

  @override
  String get name => this._name;
  @override
  int get width => this._width;
  @override
  int get poly => this._poly;
  @override
  int get init => this._init;
  @override
  bool get refIn => this._refIn;
  @override
  bool get refOut => this._refOut;
  @override
  int get xorOut => this._xorOut;

  @override
  int calculate(List<int> data) {
    // width < 8：crc 需要与 byte 左对齐
    final digits1 = max(8 - this.width, 0);
    // width > 8：byte 需要与 crc 左对齐
    final digits2 = max(this.width - 8, 0);
    var crc = this.init << digits1;
    for (var item in data) {
      if (this.refIn) {
        // 反转每个字节
        item = item.reverse8();
      }
      crc ^= item << digits2;
      final expected = 0x80 << digits2;
      final fixed = this.poly << digits1;
      for (var i = 0; i < 8; i++) {
        final actual = crc & expected;
        if (actual == expected) {
          crc = (crc << 1) ^ fixed;
        } else {
          crc <<= 1;
        }
      }
    }
    crc >>= digits1;
    if (this.refOut) {
      // 反转 crc
      crc = crc.reverse32();
      crc >>= 32 - this.width;
    }
    crc ^= this.xorOut;
    final mask = pow(2, this.width).toInt() - 1;
    crc &= mask;
    return crc;
  }

  @override
  bool verify(List<int> data, int crc) {
    final expected = this.calculate(data);
    return crc == expected;
  }
}
